
# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT

from typing import List
from typing import Optional

from langchain_core.tools import BaseTool
from langchain_mcp_adapters.client import MultiServerMCPClient


class LangChainMcpAdapter:
    """
    Adapter class to fetch tools from a Multi-Client Protocol (MCP) server and return them as
    LangChain-compatible tools. This class provides static methods for interacting with MCP servers.
    """

    @staticmethod
    async def get_mcp_tools(
        server_url: str,
        allowed_tools: Optional[List[str]] = None,
    ) -> List[BaseTool]:
        """
        Fetches tools from the given MCP server and returns them as a list of LangChain-compatible tools.

        :param server_url: URL of the MCP server, e.g. https://mcp.deepwiki.com/mcp or http://localhost:8000/mcp/
        :param allowed_tools: Optional list of tool names to filter from the server's available tools.
                              If None, all tools from the server will be returned.

        :return: A list of LangChain BaseTool instances retrieved from the MCP server.
        """

        client = MultiServerMCPClient(
            {
                "mcp_tool": {
                    "url": server_url,
                    "transport": "streamable_http",
                    # As of 09/22/2025, Neuro-SAN does not support passing authentication
                    # details (e.g., tokens) through the HOCON interface.
                    # In the future, these details will be provided in the "headers" key here.
                }
            }
        )

        # The get_tools() method returns a list of StructuredTool instances, which are subclasses of BaseTool.
        # Internally, it calls load_mcp_tools(), which uses an `async with create_session(...)` block.
        # This guarantees that any temporary MCP session created is properly closed when the block exits,
        # even if an error is raised during tool loading.
        # See: https://github.com/langchain-ai/langchain-mcp-adapters/blob/main/langchain_mcp_adapters/tools.py#L164
        # Optimization:
        #   It's possible we might want to cache these results somehow to minimize tool calls.
        mcp_tools: List[BaseTool] = await client.get_tools()

        # If allowed_tools is provided, filter the list to include only those tools.
        if allowed_tools:
            mcp_tools = [tool for tool in mcp_tools if tool.name in allowed_tools]

        for tool in mcp_tools:
            # Add "langchain_tool" tags so journal callback can idenitify it.
            # Thus MCP tools are treated as Langchain tools and can be reported in the thinking file.
            tool.tags = ["langchain_tool"]

        return mcp_tools




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT

from typing import Any
from typing import Dict

from langchain_core.language_models.base import BaseLanguageModel

from leaf_common.config.resolver import Resolver

from neuro_san.internals.run_context.langchain.llms.langchain_llm_factory import LangChainLlmFactory


class StandardLangChainLlmFactory(LangChainLlmFactory):
    """
    Factory class for LLM operations

    Most methods take a config dictionary which consists of the following keys:

        "model_name"                The name of the model.
                                    Default if not specified is "gpt-3.5-turbo"

        "temperature"               A float "temperature" value with which to
                                    initialize the chat model.  In general,
                                    higher temperatures yield more random results.
                                    Default if not specified is 0.7

        "prompt_token_fraction"     The fraction of total tokens (not necessarily words
                                    or letters) to use for a prompt. Each model_name
                                    has a documented number of max_tokens it can handle
                                    which is a total count of message + response tokens
                                    which goes into the calculation involved in
                                    get_max_prompt_tokens().
                                    By default the value is 0.5.

        "max_tokens"                The maximum number of tokens to use in
                                    get_max_prompt_tokens(). By default this comes from
                                    the model description in this class.
    """

    def create_base_chat_model(self, config: Dict[str, Any]) -> BaseLanguageModel:
        """
        Create a BaseLanguageModel from the fully-specified llm config.
        :param config: The fully specified llm config which is a product of
                    _create_full_llm_config() above.
        :return: A BaseLanguageModel (can be Chat or LLM)
                Can raise a ValueError if the config's class or model_name value is
                unknown to this method.
        """
        # Construct the LLM
        llm: BaseLanguageModel = None
        chat_class: str = config.get("class")
        if chat_class is not None:
            chat_class = chat_class.lower()

        # Check for key "model_name", "model", and "model_id" to use as model name
        # If the config is from default_llm_info, this is always "model_name"
        # but with user-specified config, it is possible to have the other keys will be specifed instead.
        model_name: str = config.get("model_name") or config.get("model") or config.get("model_id")

        # Set up a resolver to use to resolve lazy imports of classes from
        # langchain_* packages to prevent installing the world.
        resolver = Resolver()

        if chat_class == "openai":

            # OpenAI is the one chat class that we do not require any extra installs.
            # This is what we want to work out of the box.
            # Nevertheless, have it go through the same lazy-loading resolver rigamarole as the others.

            # pylint: disable=invalid-name
            ChatOpenAI = resolver.resolve_class_in_module("ChatOpenAI",
                                                          module_name="langchain_openai.chat_models.base",
                                                          install_if_missing="langchain-openai")
            llm = ChatOpenAI(
                model_name=model_name,
                temperature=config.get("temperature"),
                openai_api_key=self.get_value_or_env(config, "openai_api_key",
                                                     "OPENAI_API_KEY"),
                openai_api_base=self.get_value_or_env(config, "openai_api_base",
                                                      "OPENAI_API_BASE"),
                openai_organization=self.get_value_or_env(config, "openai_organization",
                                                          "OPENAI_ORG_ID"),
                openai_proxy=self.get_value_or_env(config, "openai_organization",
                                                   "OPENAI_PROXY"),
                request_timeout=config.get("request_timeout"),
                max_retries=config.get("max_retries"),
                presence_penalty=config.get("presence_penalty"),
                frequency_penalty=config.get("frequency_penalty"),
                seed=config.get("seed"),
                logprobs=config.get("logprobs"),
                top_logprobs=config.get("top_logprobs"),
                logit_bias=config.get("logit_bias"),
                streaming=True,     # streaming is always on. Without it token counting will not work.
                n=1,                # n is always 1.  neuro-san will only ever consider one chat completion.
                top_p=config.get("top_p"),
                max_tokens=config.get("max_tokens"),    # This is always for output
                tiktoken_model_name=config.get("tiktoken_model_name"),
                stop=config.get("stop"),

                # The following three parameters are for reasoning models only.
                reasoning=config.get("reasoning"),
                reasoning_effort=config.get("reasoning_effort"),
                verbosity=config.get("verbosity"),

                # If omitted, this defaults to the global verbose value,
                # accessible via langchain_core.globals.get_verbose():
                # https://github.com/langchain-ai/langchain/blob/master/libs/core/langchain_core/globals.py#L53
                #
                # However, accessing the global verbose value during concurrent initialization
                # can trigger the following warning:
                #
                # UserWarning: Importing verbose from langchain root module is no longer supported.
                # Please use langchain.globals.set_verbose() / langchain.globals.get_verbose() instead.
                # old_verbose = langchain.verbose
                #
                # To prevent this, we explicitly set verbose=False here (which matches the default
                # global verbose value) so that the warning is never triggered.
                verbose=False,

                # Set stream_usage to True in order to get token counting chunks.
                stream_usage=True
            )
        elif chat_class == "azure-openai":
            model_kwargs: Dict[str, Any] = {
                "stream_options": {
                    "include_usage": True
                }
            }
            openai_api_key: str = self.get_value_or_env(config, "openai_api_key", "AZURE_OPENAI_API_KEY")
            if openai_api_key is None:
                openai_api_key = self.get_value_or_env(config, "openai_api_key", "OPENAI_API_KEY")

            # AzureChatOpenAI just happens to come with langchain_openai
            # pylint: disable=invalid-name
            AzureChatOpenAI = resolver.resolve_class_in_module("AzureChatOpenAI",
                                                               module_name="langchain_openai.chat_models.azure",
                                                               install_if_missing="langchain-openai")
            llm = AzureChatOpenAI(
                model_name=model_name,
                temperature=config.get("temperature"),
                openai_api_key=openai_api_key,
                openai_api_base=self.get_value_or_env(config, "openai_api_base",
                                                      "OPENAI_API_BASE"),
                openai_organization=self.get_value_or_env(config, "openai_organization",
                                                          "OPENAI_ORG_ID"),
                openai_proxy=self.get_value_or_env(config, "openai_organization",
                                                   "OPENAI_PROXY"),
                request_timeout=config.get("request_timeout"),
                max_retries=config.get("max_retries"),
                presence_penalty=config.get("presence_penalty"),
                frequency_penalty=config.get("frequency_penalty"),
                seed=config.get("seed"),
                logprobs=config.get("logprobs"),
                top_logprobs=config.get("top_logprobs"),
                logit_bias=config.get("logit_bias"),
                streaming=True,     # streaming is always on. Without it token counting will not work.
                n=1,                # n is always 1.  neuro-san will only ever consider one chat completion.
                top_p=config.get("top_p"),
                max_tokens=config.get("max_tokens"),    # This is always for output
                tiktoken_model_name=config.get("tiktoken_model_name"),
                stop=config.get("stop"),

                # If omitted, this defaults to the global verbose value,
                # accessible via langchain_core.globals.get_verbose():
                # https://github.com/langchain-ai/langchain/blob/master/libs/core/langchain_core/globals.py#L53
                #
                # However, accessing the global verbose value during concurrent initialization
                # can trigger the following warning:
                #
                # UserWarning: Importing verbose from langchain root module is no longer supported.
                # Please use langchain.globals.set_verbose() / langchain.globals.get_verbose() instead.
                # old_verbose = langchain.verbose
                #
                # To prevent this, we explicitly set verbose=False here (which matches the default
                # global verbose value) so that the warning is never triggered.
                verbose=False,

                # Azure-specific
                azure_endpoint=self.get_value_or_env(config, "azure_endpoint",
                                                     "AZURE_OPENAI_ENDPOINT"),
                deployment_name=self.get_value_or_env(config, "deployment_name",
                                                      "AZURE_OPENAI_DEPLOYMENT_NAME"),
                openai_api_version=self.get_value_or_env(config, "openai_api_version",
                                                         "OPENAI_API_VERSION"),

                # AD here means "ActiveDirectory"
                azure_ad_token=self.get_value_or_env(config, "azure_ad_token",
                                                     "AZURE_OPENAI_AD_TOKEN"),
                model_version=config.get("model_version"),
                openai_api_type=self.get_value_or_env(config, "openai_api_type",
                                                      "OPENAI_API_TYPE"),
                # Needed for token counting
                model_kwargs=model_kwargs,
            )
        elif chat_class == "anthropic":

            # Use lazy loading to prevent installing the world
            # pylint: disable=invalid-name
            ChatAnthropic = resolver.resolve_class_in_module("ChatAnthropic",
                                                             module_name="langchain_anthropic.chat_models",
                                                             install_if_missing="langchain-anthropic")
            llm = ChatAnthropic(
                model_name=model_name,
                max_tokens=config.get("max_tokens"),    # This is always for output
                temperature=config.get("temperature"),
                top_k=config.get("top_k"),
                top_p=config.get("top_p"),
                default_request_timeout=config.get("default_request_timeout"),
                max_retries=config.get("max_retries"),
                stop_sequences=config.get("stop_sequences"),
                anthropic_api_url=self.get_value_or_env(config, "anthropic_api_url",
                                                        "ANTHROPIC_API_URL"),
                anthropic_api_key=self.get_value_or_env(config, "anthropic_api_key",
                                                        "ANTHROPIC_API_KEY"),
                streaming=True,     # streaming is always on. Without it token counting will not work.
                # Set stream_usage to True in order to get token counting chunks.
                stream_usage=True,

                # If omitted, this defaults to the global verbose value,
                # accessible via langchain_core.globals.get_verbose():
                # https://github.com/langchain-ai/langchain/blob/master/libs/core/langchain_core/globals.py#L53
                #
                # However, accessing the global verbose value during concurrent initialization
                # can trigger the following warning:
                #
                # UserWarning: Importing verbose from langchain root module is no longer supported.
                # Please use langchain.globals.set_verbose() / langchain.globals.get_verbose() instead.
                # old_verbose = langchain.verbose
                #
                # To prevent this, we explicitly set verbose=False here (which matches the default
                # global verbose value) so that the warning is never triggered.
                verbose=False,
            )
        elif chat_class == "ollama":

            # Use lazy loading to prevent installing the world
            # pylint: disable=invalid-name
            ChatOllama = resolver.resolve_class_in_module("ChatOllama",
                                                          module_name="langchain_ollama",
                                                          install_if_missing="langchain-ollama")
            # Higher temperature is more random
            llm = ChatOllama(
                model=model_name,
                mirostat=config.get("mirostat"),
                mirostat_eta=config.get("mirostat_eta"),
                mirostat_tau=config.get("mirostat_tau"),
                num_ctx=config.get("num_ctx"),
                num_gpu=config.get("num_gpu"),
                num_thread=config.get("num_thread"),
                num_predict=config.get("num_predict", config.get("max_tokens")),
                reasoning=config.get("reasoning"),
                repeat_last_n=config.get("repeat_last_n"),
                repeat_penalty=config.get("repeat_penalty"),
                temperature=config.get("temperature"),
                seed=config.get("seed"),
                stop=config.get("stop"),
                tfs_z=config.get("tfs_z"),
                top_k=config.get("top_k"),
                top_p=config.get("top_p"),
                keep_alive=config.get("keep_alive"),
                base_url=config.get("base_url"),

                # If omitted, this defaults to the global verbose value,
                # accessible via langchain_core.globals.get_verbose():
                # https://github.com/langchain-ai/langchain/blob/master/libs/core/langchain_core/globals.py#L53
                #
                # However, accessing the global verbose value during concurrent initialization
                # can trigger the following warning:
                #
                # UserWarning: Importing verbose from langchain root module is no longer supported.
                # Please use langchain.globals.set_verbose() / langchain.globals.get_verbose() instead.
                # old_verbose = langchain.verbose
                #
                # To prevent this, we explicitly set verbose=False here (which matches the default
                # global verbose value) so that the warning is never triggered.
                verbose=False,
            )
        elif chat_class == "nvidia":

            # Use lazy loading to prevent installing the world
            # pylint: disable=invalid-name
            ChatNVIDIA = resolver.resolve_class_in_module("ChatNVIDIA",
                                                          module_name="langchain_nvidia_ai_endpoints",
                                                          install_if_missing="langchain-nvidia-ai-endpoints")
            # Higher temperature is more random
            llm = ChatNVIDIA(
                base_url=config.get("base_url"),
                model=model_name,
                temperature=config.get("temperature"),
                max_tokens=config.get("max_tokens"),
                top_p=config.get("top_p"),
                seed=config.get("seed"),
                stop=config.get("stop"),

                # If omitted, this defaults to the global verbose value,
                # accessible via langchain_core.globals.get_verbose():
                # https://github.com/langchain-ai/langchain/blob/master/libs/core/langchain_core/globals.py#L53
                #
                # However, accessing the global verbose value during concurrent initialization
                # can trigger the following warning:
                #
                # UserWarning: Importing verbose from langchain root module is no longer supported.
                # Please use langchain.globals.set_verbose() / langchain.globals.get_verbose() instead.
                # old_verbose = langchain.verbose
                #
                # To prevent this, we explicitly set verbose=False here (which matches the default
                # global verbose value) so that the warning is never triggered.
                verbose=False,
                nvidia_api_key=self.get_value_or_env(config, "nvidia_api_key",
                                                     "NVIDIA_API_KEY"),
                nvidia_base_url=self.get_value_or_env(config, "nvidia_base_url",
                                                      "NVIDIA_BASE_URL"),
            )
        elif chat_class == "gemini":

            # Use lazy loading to prevent installing the world
            # pylint: disable=invalid-name
            ChatGoogleGenerativeAI = resolver.resolve_class_in_module("ChatGoogleGenerativeAI",
                                                                      module_name="langchain_google_genai.chat_models",
                                                                      install_if_missing="langchain-google-genai")
            llm = ChatGoogleGenerativeAI(
                model=model_name,
                google_api_key=self.get_value_or_env(config, "google_api_key",
                                                     "GOOGLE_API_KEY"),
                max_retries=config.get("max_retries"),
                max_tokens=config.get("max_tokens"),    # This is always for output
                n=config.get("n"),
                temperature=config.get("temperature"),
                timeout=config.get("timeout"),
                top_k=config.get("top_k"),
                top_p=config.get("top_p"),

                # If omitted, this defaults to the global verbose value,
                # accessible via langchain_core.globals.get_verbose():
                # https://github.com/langchain-ai/langchain/blob/master/libs/core/langchain_core/globals.py#L53
                #
                # However, accessing the global verbose value during concurrent initialization
                # can trigger the following warning:
                #
                # UserWarning: Importing verbose from langchain root module is no longer supported.
                # Please use langchain.globals.set_verbose() / langchain.globals.get_verbose() instead.
                # old_verbose = langchain.verbose
                #
                # To prevent this, we explicitly set verbose=False here (which matches the default
                # global verbose value) so that the warning is never triggered.
                verbose=False,
            )
        elif chat_class == "bedrock":

            # Use lazy loading to prevent installing the world
            # pylint: disable=invalid-name
            ChatBedrock = resolver.resolve_class_in_module("ChatBedrock",
                                                           module_name="langchain_aws",
                                                           install_if_missing="langchain-aws")
            llm = ChatBedrock(
                model=model_name,
                aws_access_key_id=self.get_value_or_env(config, "aws_access_key_id", "AWS_ACCESS_KEY_ID"),
                aws_secret_access_key=self.get_value_or_env(config, "aws_secret_access_key", "AWS_SECRET_ACCESS_KEY"),
                aws_session_token=self.get_value_or_env(config, "aws_session_token", "AWS_SESSION_TOKEN"),
                base_model_id=config.get("base_model_id"),
                beta_use_converse_api=config.get("beta_use_converse_api"),
                cache=config.get("cache"),
                config=config.get("config"),
                credentials_profile_name=config.get("credentials_profile_name"),
                custom_get_token_ids=config.get("custom_get_token_ids"),
                endpoint_url=config.get("endpoint_url"),
                guardrails=config.get("guardrails"),
                max_tokens=config.get("max_tokens"),
                metadata=config.get("metadata"),
                provider=config.get("provider"),
                rate_limiter=config.get("rate_limiter"),
                region_name=config.get("region_name"),
                stop_sequences=config.get("stop_sequences"),
                streaming=True,
                system_prompt_with_tools=config.get("system_prompt_with_tools"),
                tags=config.get("tags"),
                temperature=config.get("temperature"),

                # If omitted, this defaults to the global verbose value,
                # accessible via langchain_core.globals.get_verbose():
                # https://github.com/langchain-ai/langchain/blob/master/libs/core/langchain_core/globals.py#L53
                #
                # However, accessing the global verbose value during concurrent initialization
                # can trigger the following warning:
                #
                # UserWarning: Importing verbose from langchain root module is no longer supported.
                # Please use langchain.globals.set_verbose() / langchain.globals.get_verbose() instead.
                # old_verbose = langchain.verbose
                #
                # To prevent this, we explicitly set verbose=False here (which matches the default
                # global verbose value) so that the warning is never triggered.
                verbose=False,
            )
        elif chat_class is None:
            raise ValueError(f"Class name {chat_class} for model_name {model_name} is unspecified.")
        else:
            raise ValueError(f"Class {chat_class} for model_name {model_name} is unrecognized.")

        return llm




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT
from typing import Any
from typing import Dict
from typing import List

import os
import threading

from argparse import ArgumentParser

from leaf_server_common.server.server_loop_callbacks import ServerLoopCallbacks
from leaf_server_common.logging.logging_setup import setup_logging

from neuro_san import DEPLOY_DIR
from neuro_san import TOP_LEVEL_DIR
from neuro_san.interfaces.agent_session import AgentSession
from neuro_san.internals.graph.persistence.registry_manifest_restorer import RegistryManifestRestorer
from neuro_san.internals.graph.registry.agent_network import AgentNetwork
from neuro_san.internals.network_providers.agent_network_storage import AgentNetworkStorage
from neuro_san.service.grpc.grpc_agent_server import DEFAULT_SERVER_NAME
from neuro_san.service.grpc.grpc_agent_server import DEFAULT_SERVER_NAME_FOR_LOGS
from neuro_san.service.grpc.grpc_agent_server import DEFAULT_MAX_CONCURRENT_REQUESTS
from neuro_san.service.grpc.grpc_agent_server import DEFAULT_REQUEST_LIMIT
from neuro_san.service.http.config.http_server_config import DEFAULT_HTTP_CONNECTIONS_BACKLOG
from neuro_san.service.http.config.http_server_config import DEFAULT_HTTP_IDLE_CONNECTIONS_TIMEOUT_SECONDS
from neuro_san.service.http.config.http_server_config import DEFAULT_HTTP_SERVER_INSTANCES
from neuro_san.service.http.config.http_server_config import DEFAULT_HTTP_SERVER_MONITOR_INTERVAL_SECONDS
from neuro_san.service.http.config.http_server_config import HttpServerConfig
from neuro_san.service.grpc.grpc_agent_server import GrpcAgentServer
from neuro_san.service.grpc.grpc_agent_service import GrpcAgentService
from neuro_san.service.http.server.http_server import HttpServer
from neuro_san.service.watcher.main_loop.storage_watcher import StorageWatcher
from neuro_san.service.utils.server_status import ServerStatus
from neuro_san.service.utils.server_context import ServerContext


# pylint: disable=too-many-instance-attributes
class ServerMainLoop(ServerLoopCallbacks):
    """
    This class handles the service main loop.
    """

    def __init__(self):
        """
        Constructor
        """
        self.grpc_port: int = 0
        self.http_port: int = 0

        self.agent_networks: Dict[str, AgentNetwork] = {}

        self.server_name: str = DEFAULT_SERVER_NAME
        self.server_name_for_logs: str = DEFAULT_SERVER_NAME_FOR_LOGS
        self.max_concurrent_requests: int = DEFAULT_MAX_CONCURRENT_REQUESTS
        self.request_limit: int = DEFAULT_REQUEST_LIMIT
        self.forwarded_request_metadata: str = GrpcAgentServer.DEFAULT_FORWARDED_REQUEST_METADATA
        self.usage_logger_metadata: str = ""
        self.service_openapi_spec_file: str = self._get_default_openapi_spec_path()
        self.grpc_server: GrpcAgentServer = None
        self.http_server: HttpServer = None
        self.server_context = ServerContext()
        self.http_server_config = HttpServerConfig()
        self.watcher_config: Dict[str, Any] = {}

    def prepare_args(self) -> ArgumentParser:
        """
        :return: An ArgumentParser set up to parse command-line arguments
        """
        # Set up the CLI parser
        arg_parser = ArgumentParser()

        arg_parser.add_argument("--port", type=int,
                                default=int(os.environ.get("AGENT_PORT", AgentSession.DEFAULT_PORT)),
                                help="Port number for the grpc service")
        arg_parser.add_argument("--http_port", type=int,
                                default=int(os.environ.get("AGENT_HTTP_PORT", AgentSession.DEFAULT_HTTP_PORT)),
                                help="Port number for http service endpoint")
        arg_parser.add_argument("--server_name", type=str,
                                default=str(os.environ.get("AGENT_SERVER_NAME", self.server_name)),
                                help="Name of the service for health reporting purposes.")
        arg_parser.add_argument("--server_name_for_logs", type=str,
                                default=str(os.environ.get("AGENT_SERVER_NAME_FOR_LOGS", self.server_name_for_logs)),
                                help="Name of the service as seen in logs")
        arg_parser.add_argument("--max_concurrent_requests", type=int,
                                default=int(os.environ.get("AGENT_MAX_CONCURRENT_REQUESTS",
                                                           self.max_concurrent_requests)),
                                help="Maximum number of requests that can be served at the same time")
        arg_parser.add_argument("--request_limit", type=int,
                                default=int(os.environ.get("AGENT_REQUEST_LIMIT", self.request_limit)),
                                help="Number of requests served before the server shuts down in an orderly fashion")
        arg_parser.add_argument("--forwarded_request_metadata", type=str,
                                default=os.environ.get("AGENT_FORWARDED_REQUEST_METADATA",
                                                       self.forwarded_request_metadata),
                                help="Space-delimited list of http request metadata keys to forward "
                                     "to logs/other requests")
        arg_parser.add_argument("--usage_logger_metadata", type=str,
                                default=os.environ.get("AGENT_USAGE_LOGGER_METADATA", ""),
                                help="Space-delimited list of http request metadata keys to forward "
                                     "to models usage statistics logger")
        arg_parser.add_argument("--openapi_service_spec_path", type=str,
                                default=os.environ.get("AGENT_OPENAPI_SPEC",
                                                       self.service_openapi_spec_file),
                                help="File path to OpenAPI service specification document.")
        arg_parser.add_argument("--manifest_update_period_seconds", type=int,
                                default=int(os.environ.get("AGENT_MANIFEST_UPDATE_PERIOD_SECONDS", "0")),
                                help="Periodic run-time update period for manifest in seconds."
                                     " Value <= 0 disables updates.")
        arg_parser.add_argument("--http_connections_backlog", type=int,
                                default=int(os.environ.get("AGENT_HTTP_CONNECTIONS_BACKLOG",
                                                           DEFAULT_HTTP_CONNECTIONS_BACKLOG)),
                                help="Size of backlog for TCP connections to http server.")
        arg_parser.add_argument("--http_idle_connections_timeout", type=int,
                                default=int(os.environ.get("AGENT_HTTP_IDLE_CONNECTIONS_TIMEOUT",
                                                           DEFAULT_HTTP_IDLE_CONNECTIONS_TIMEOUT_SECONDS)),
                                help="Timeout in seconds before idle and alive connection to http server"
                                     "will be closed")
        arg_parser.add_argument("--http_server_instances", type=int,
                                default=int(os.environ.get("AGENT_HTTP_SERVER_INSTANCES",
                                                           DEFAULT_HTTP_SERVER_INSTANCES)),
                                help="Number of http server instances to be created "
                                     "one instance per separate process")
        arg_parser.add_argument("--http_resources_monitor_interval_seconds", type=int,
                                default=int(os.environ.get("AGENT_HTTP_RESOURCES_MONITOR_INTERVAL",
                                                           DEFAULT_HTTP_SERVER_MONITOR_INTERVAL_SECONDS)),
                                help="Http server resources monitoring/logging interval in seconds "
                                     "0 means no logging")
        return arg_parser

    def parse_args(self):
        """
        Parse command-line arguments into member variables
        """
        arg_parser: ArgumentParser = self.prepare_args()

        # Actually parse the args into class variables

        # Incorrectly flagged as Path Traversal 3, 7
        # See destination below ~ line 139, 154 for explanation.
        args = arg_parser.parse_args()
        self.server_name = args.server_name
        server_status = ServerStatus(self.server_name)
        self.server_context.set_server_status(server_status)

        self.grpc_port = args.port
        if self.grpc_port == 0:
            server_status.grpc_service.set_requested(False)
        self.http_port = args.http_port
        if self.http_port == 0:
            server_status.http_service.set_requested(False)

        self.server_name_for_logs = args.server_name_for_logs
        self.max_concurrent_requests = args.max_concurrent_requests
        self.request_limit = args.request_limit
        self.forwarded_request_metadata = args.forwarded_request_metadata
        if not self.forwarded_request_metadata:
            self.forwarded_request_metadata = ""
        self.usage_logger_metadata = args.usage_logger_metadata
        if not self.usage_logger_metadata:
            self.usage_logger_metadata = self.forwarded_request_metadata
        self.service_openapi_spec_file = args.openapi_service_spec_path
        if args.manifest_update_period_seconds <= 0:
            # StorageWatcher is disabled:
            server_status.updater.set_requested(False)

        self.http_server_config.http_connections_backlog = args.http_connections_backlog
        self.http_server_config.http_idle_connection_timeout_seconds = args.http_idle_connections_timeout
        self.http_server_config.http_server_instances = args.http_server_instances
        self.http_server_config.http_server_monitor_interval_seconds = args.http_resources_monitor_interval_seconds
        self.http_server_config.http_port = args.http_port

        manifest_restorer = RegistryManifestRestorer()
        manifest_agent_networks: Dict[str, AgentNetwork] = manifest_restorer.restore()
        manifest_files: List[str] = manifest_restorer.get_manifest_files()

        self.watcher_config = {
            "manifest_path": manifest_files[0],    # For now, only one
            "manifest_update_period_seconds": args.manifest_update_period_seconds,
        }

        self.agent_networks = manifest_agent_networks

    def _get_default_openapi_spec_path(self) -> str:
        """
        Return a file path to default location of OpenAPI specification file
        for neuro-san service.
        """
        return TOP_LEVEL_DIR.get_file_in_basis("api/grpc/agent_service.json")

    def main_loop(self):
        """
        Command line entry point
        """
        self.parse_args()

        # Make for easy running from the neuro-san repo
        if os.environ.get("AGENT_SERVICE_LOG_JSON") is None:
            # Use the log file that is local to the repo
            os.environ["AGENT_SERVICE_LOG_JSON"] = DEPLOY_DIR.get_file_in_basis("logging.json")

        # Construct forwarded metadata list as a union of
        # self.forwarded_request_metadata and self.usage_logger_metadata
        metadata_set = set(self.forwarded_request_metadata.split())
        metadata_set = metadata_set | set(self.usage_logger_metadata.split())
        metadata_str: str = " ".join(sorted(metadata_set))

        server_status: ServerStatus = self.server_context.get_server_status()

        if server_status.grpc_service.is_requested():
            self.grpc_server = GrpcAgentServer(
                self.grpc_port,
                server_loop_callbacks=self,
                server_context=self.server_context,
                server_name=self.server_name,
                server_name_for_logs=self.server_name_for_logs,
                max_concurrent_requests=self.max_concurrent_requests,
                request_limit=self.request_limit,
                forwarded_request_metadata=metadata_str)
            self.grpc_server.prepare_for_serving()

        if server_status.updater.is_requested():
            if not server_status.grpc_service.is_requested():
                current_dir: str = os.path.dirname(os.path.abspath(__file__))
                setup_logging(server_status.updater.get_service_name(),
                              current_dir,
                              'AGENT_SERVICE_LOG_JSON',
                              'AGENT_SERVICE_LOG_LEVEL')
            watcher = StorageWatcher(self.watcher_config, self.server_context)
            watcher.start()

        if server_status.http_service.is_requested():
            # Create HTTP server;
            self.http_server = HttpServer(
                self.server_context,
                self.http_server_config,
                self.service_openapi_spec_file,
                self.request_limit,
                forwarded_request_metadata=metadata_str)

        # Now - our servers (gRPC and http) are created and listen to updates of network_storage
        # Perform the initial setup
        network_storage_dict: Dict[str, AgentNetworkStorage] = self.server_context.get_network_storage_dict()
        public_storage: AgentNetworkStorage = network_storage_dict.get("public")
        public_storage.setup_agent_networks(self.agent_networks)

        # Start all services:
        http_server_thread = None
        if server_status.http_service.is_requested():
            http_server_thread = threading.Thread(target=self.http_server, args=(self.grpc_server,), daemon=True)
            http_server_thread.start()

        if server_status.grpc_service.is_requested():
            self.grpc_server.serve()

        if http_server_thread is not None:
            http_server_thread.join()

    def loop_callback(self) -> bool:
        """
        Periodically called by the main server loop of ServerLifetime.
        """
        # Report back on service activity so the ServerLifetime that calls
        # this method can properly yield/sleep depending on how many requests
        # are in motion.
        agent_services: List[GrpcAgentService] = self.grpc_server.get_services()
        for agent_service in agent_services:
            if agent_service.get_request_count() > 0:
                return True

        return False


if __name__ == '__main__':
    ServerMainLoop().main_loop()




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT
"""
See class comment for details
"""
from typing import Any
from typing import Dict
from typing import Generator

import json

from neuro_san.service.generic.async_agent_service import AsyncAgentService
from neuro_san.service.http.handlers.base_request_handler import BaseRequestHandler


class StreamingChatHandler(BaseRequestHandler):
    """
    Handler class for neuro-san streaming chat API call.
    """

    async def stream_out(self,
                         generator: Generator[Dict[str, Any], None, None]) -> int:
        """
        Process streaming out generator output to HTTP connection.
        :param generator: async chat response generator
        :return: number of chat responses streamed out.
        """
        # Set up headers for chunked response
        self.set_header("Content-Type", "application/json-lines")
        self.set_header("Transfer-Encoding", "chunked")
        # Flush headers immediately
        flush_ok: bool = await self.do_flush()
        if not flush_ok:
            return 0

        sent_out: int = 0
        async for result_dict in generator:
            result_str: str = json.dumps(result_dict) + "\n"
            self.write(result_str)
            flush_ok = await self.do_flush()
            if not flush_ok:
                return sent_out
            sent_out += 1
        return sent_out

    async def post(self, agent_name: str):
        """
        Implementation of POST request handler for streaming chat API call.
        """

        metadata: Dict[str, Any] = self.get_metadata()
        service: AsyncAgentService = await self.get_service(agent_name, metadata)
        if service is None:
            return

        self.application.start_client_request(metadata, f"{agent_name}/streaming_chat")
        try:
            # Parse JSON body
            data = json.loads(self.request.body)
            result_generator = service.streaming_chat(data, metadata)
            await self.stream_out(result_generator)

        except Exception as exc:  # pylint: disable=broad-exception-caught
            self.process_exception(exc)
        finally:
            # We are done with response stream:
            self.do_finish()
            self.application.finish_client_request(metadata, f"{agent_name}/streaming_chat", get_stats=True)




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT
from typing import Any
from typing import Dict

from leaf_common.time.timeout import Timeout
from leaf_common.asyncio.asyncio_executor_pool import AsyncioExecutorPool

from neuro_san.client.direct_agent_storage_util import DirectAgentStorageUtil
from neuro_san.interfaces.agent_session import AgentSession
from neuro_san.internals.interfaces.context_type_toolbox_factory import ContextTypeToolboxFactory
from neuro_san.internals.graph.registry.agent_network import AgentNetwork
from neuro_san.internals.interfaces.context_type_llm_factory import ContextTypeLlmFactory
from neuro_san.internals.run_context.factory.master_toolbox_factory import MasterToolboxFactory
from neuro_san.internals.run_context.factory.master_llm_factory import MasterLlmFactory
from neuro_san.internals.graph.persistence.agent_network_restorer import AgentNetworkRestorer
from neuro_san.internals.interfaces.agent_network_provider import AgentNetworkProvider
from neuro_san.internals.network_providers.agent_network_storage import AgentNetworkStorage
from neuro_san.session.direct_agent_session import DirectAgentSession
from neuro_san.session.external_agent_session_factory import ExternalAgentSessionFactory
from neuro_san.session.missing_agent_check import MissingAgentCheck
from neuro_san.session.session_invocation_context import SessionInvocationContext


class DirectAgentSessionFactory:
    """
    Sets up everything needed to use a DirectAgentSession more as a library.
    This includes:
        * Some reading of AgentNetworks
        * Setting up AgentNetworkStorage with agent networks
          which were read in
        * Initializing an LlmFactory
    """

    def __init__(self):
        """
        Constructor
        """
        self.network_storage: AgentNetworkStorage = DirectAgentStorageUtil.create_network_storage()

    def create_session(self, agent_name: str, use_direct: bool = False,
                       metadata: Dict[str, str] = None, umbrella_timeout: Timeout = None) -> AgentSession:
        """
        :param agent_name: The name of the agent to use for the session.
                This name can be something in the manifest file (with no file suffix)
                or a specific full-reference to an agent network's hocon file.
        :param use_direct: When True, will use a Direct session for
                    external agents that would reside on the same server.
        :param metadata: A grpc metadata of key/value pairs to be inserted into
                         the header. Default is None. Preferred format is a
                         dictionary of string keys to string values.
        :param umbrella_timeout: A Timeout object to periodically check in loops.
                        Default is None (no timeout).
        """

        agent_network: AgentNetwork = self.get_agent_network(agent_name)
        config: Dict[str, Any] = agent_network.get_config()
        llm_factory: ContextTypeLlmFactory = MasterLlmFactory.create_llm_factory(config)
        toolbox_factory: ContextTypeToolboxFactory = MasterToolboxFactory.create_toolbox_factory(config)
        # Load once now that we know what tool registry to use.
        llm_factory.load()
        toolbox_factory.load()

        factory = ExternalAgentSessionFactory(use_direct=use_direct, network_storage=self.network_storage)
        executors_pool: AsyncioExecutorPool = AsyncioExecutorPool()
        invocation_context = SessionInvocationContext(factory, executors_pool, llm_factory, toolbox_factory, metadata)
        invocation_context.start()
        session: DirectAgentSession = DirectAgentSession(agent_network=agent_network,
                                                         invocation_context=invocation_context,
                                                         metadata=metadata,
                                                         umbrella_timeout=umbrella_timeout)
        return session

    def get_agent_network(self, agent_name: str) -> AgentNetwork:
        """
        :param agent_name: The name of the agent whose AgentNetwork we want to get.
                This name can be something in the manifest file (with no file suffix)
                or a specific full-reference to an agent network's hocon file.
        :return: The AgentNetwork corresponding to that agent.
        """

        if agent_name is None or len(agent_name) == 0:
            return None

        agent_network: AgentNetwork = None
        if agent_name.endswith(".hocon") or agent_name.endswith(".json"):
            # We got a specific file name
            restorer = AgentNetworkRestorer()
            agent_network = restorer.restore(file_reference=agent_name)
        else:
            # Use the standard stuff available via the manifest file.
            agent_network_provider: AgentNetworkProvider =\
                self.network_storage.get_agent_network_provider(agent_name)
            agent_network = agent_network_provider.get_agent_network()

        # Common place for nice error messages when networks are not found
        MissingAgentCheck.check_agent_network(agent_network, agent_name)

        return agent_network




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT
from typing import Any
from typing import Dict
from typing import List
from typing import Sequence
from typing import Union

import os
import json
import logging

from pyparsing.exceptions import ParseException
from pyparsing.exceptions import ParseSyntaxException

from leaf_common.config.file_of_class import FileOfClass
from leaf_common.persistence.easy.easy_hocon_persistence import EasyHoconPersistence
from leaf_common.persistence.interface.restorer import Restorer

from neuro_san import REGISTRIES_DIR
from neuro_san.internals.interfaces.agent_name_mapper import AgentNameMapper
from neuro_san.internals.graph.persistence.agent_filetree_mapper import AgentFileTreeMapper
from neuro_san.internals.graph.persistence.agent_network_restorer import AgentNetworkRestorer
from neuro_san.internals.graph.registry.agent_network import AgentNetwork


class RegistryManifestRestorer(Restorer):
    """
    Implementation of the Restorer interface that reads the manifest file
    for agent networks/registries.
    """

    def __init__(self, manifest_files: Union[str, List[str]] = None, agent_mapper: AgentNameMapper = None):
        """
        Constructor

        :param manifest_files: Either:
            * A single local name for the manifest file listing the agents to host.
            * A list of local names for multiple manifest files to host
            * None (the default) which gets a single manifest file from a known source.
        :param agent_mapper: optional AgentNameMapper;
            if None, AgentFileTreeMapper instance will be used.
        """
        self.agent_mapper = agent_mapper
        if not self.agent_mapper:
            self.agent_mapper = AgentFileTreeMapper()

        self.manifest_files: List[str] = []

        if manifest_files is None:
            # We have no manifest list coming in, so check an env variable for a definition.
            manifest_file: str = os.environ.get("AGENT_MANIFEST_FILE")
            if manifest_file is None:
                # No env var, so fallback to whatis coded in this repo.
                manifest_file = REGISTRIES_DIR.get_file_in_basis("manifest.hocon")

            # Add what was found above
            self.manifest_files.append(manifest_file)
        elif isinstance(manifest_files, str):
            self.manifest_files.append(manifest_files)
        else:
            self.manifest_files = manifest_files

        self.logger = logging.getLogger(self.__class__.__name__)

    # pylint: disable=too-many-locals
    def restore_from_files(self, file_references: Sequence[str]) -> Dict[str, AgentNetwork]:
        """
        :param file_references: The sequence of file references to use when restoring.
        :return: a built map of agent networks
        """

        agent_networks: Dict[str, AgentNetwork] = {}

        # Loop through all the manifest files in the list to make a composite
        for manifest_file in file_references:

            one_manifest: Dict[str, Any] = {}
            if manifest_file.endswith(".hocon"):
                hocon = EasyHoconPersistence()
                try:
                    one_manifest = hocon.restore(file_reference=manifest_file)
                except (ParseException, ParseSyntaxException) as exception:
                    message: str = f"""
There was an error parsing the agent network manifest file "{manifest_file}".
See the accompanying ParseException (above) for clues as to what might be
syntactically incorrect in that file.
"""
                    raise ParseException(message) from exception
            else:
                try:
                    with open(manifest_file, "r", encoding="utf-8") as json_file:
                        one_manifest = json.load(json_file)
                except FileNotFoundError:
                    # Use the common verbiage below
                    one_manifest = None
                except json.decoder.JSONDecodeError as exception:
                    message: str = f"""
There was an error parsing the agent network manifest file "{manifest_file}".
See the accompanying JSONDecodeError exception (above) for clues as to what might be
syntactically incorrect in that file.
"""
                    raise ParseException(message) from exception

            if one_manifest is None:
                message = f"Could not find manifest file at path: {manifest_file}.\n" + """
Some common problems include:
* The file itself simply does not exist.
* Path is not an absolute path and you are invoking the server from a place
  where the path is not reachable.
* The path has a typo in it.

Double-check the value of the AGENT_MANIFEST_FILE env var and
your current working directory (pwd).
"""
                raise FileNotFoundError(message)

            for key, value in one_manifest.items():
                if not bool(value):
                    # Fast out
                    continue

                # Key here is an agent name in a form that we chose,
                # and we'll need to use an agent mapper to get to this agent definition file.
                # Keys sometimes come with quotes.
                use_key: str = key.replace(r'"', "")
                use_key = use_key.strip()
                agent_filepath: str = self.agent_mapper.agent_name_to_filepath(use_key)

                file_of_class = FileOfClass(manifest_file)
                manifest_dir: str = file_of_class.get_basis()
                registry_restorer = AgentNetworkRestorer(registry_dir=manifest_dir, agent_mapper=self.agent_mapper)
                try:
                    agent_network: AgentNetwork = registry_restorer.restore(file_reference=agent_filepath)
                except FileNotFoundError as exc:
                    self.logger.error("Failed to restore registry item %s - %s", use_key, str(exc))
                    agent_network = None
                if agent_network is not None:
                    network_name: str = self.agent_mapper.filepath_to_agent_network_name(agent_filepath)
                    agent_networks[network_name] = agent_network
                else:
                    self.logger.error("manifest registry %s not found in %s", use_key, manifest_file)

        return agent_networks

    # pylint: disable=too-many-locals
    def restore(self, file_reference: str = None) -> Dict[str, AgentNetwork]:
        """
        :param file_reference: The file reference to use when restoring.
                Default is None, implying the file reference is up to the
                implementation.
        :return: a built map of agent networks
        """
        if file_reference is not None:
            return self.restore_from_files([file_reference])

        agent_networks: Dict[str, AgentNetwork] =\
            self.restore_from_files(self.manifest_files)
        return agent_networks

    def get_manifest_files(self) -> List[str]:
        """
        Return current list of manifest files.
        """
        return self.manifest_files




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT

from typing import Any
from typing import Dict
from typing import Generator

from neuro_san.interfaces.agent_session_constants import AgentSessionConstants


class AgentSession(AgentSessionConstants):
    """
    Interface for the initiation and continuity of a single Agent session.
    """

    def function(self, request_dict: Dict[str, Any]) -> Dict[str, Any]:
        """
        :param request_dict: A dictionary version of the FunctionRequest
                    protobufs structure. Has the following keys:
                        <None>
        :return: A dictionary version of the FunctionResponse
                    protobufs structure. Has the following keys:
                "function" - the dictionary description of the function
        """
        raise NotImplementedError

    def connectivity(self, request_dict: Dict[str, Any]) -> Dict[str, Any]:
        """
        :param request_dict: A dictionary version of the ConnectivityRequest
                    protobufs structure. Has the following keys:
                        <None>
        :return: A dictionary version of the ConnectivityResponse
                    protobufs structure. Has the following keys:
                "connectivity_info" - the list of connectivity descriptions for
                                    each node in the agent network the service
                                    wants the client ot know about.
        """
        raise NotImplementedError

    def streaming_chat(self, request_dict: Dict[str, Any]) -> Generator[Dict[str, Any], None, None]:
        """
        :param request_dict: A dictionary version of the ChatRequest
                    protobufs structure. Has the following keys:
            "user_message" - A ChatMessage dict representing the user input to the chat stream
            "chat_context" - A ChatContext dict representing the state of the previous conversation
                            (if any)

        :return: An iterator of dictionary versions of the ChatResponse
                    protobufs structure. Has the following keys:
            "response"      - An optional ChatMessage dictionary.  See chat.proto for details.

            Note that responses to the chat input might be numerous and will come as they
            are produced until the system decides there are no more messages to be sent.
        """
        raise NotImplementedError
